#include "common/tcpip_info.h"
#include <cstring>
#include <boost/lexical_cast.hpp>
#include <boost/format.hpp>
#ifdef _WIN32
#include <ws2tcpip.h>
#else
#include <arpa/inet.h>
#endif

tcpipInfo::tcpipInfo() 
{
    reset();
}

void tcpipInfo::reset()
{
    data.clear();
    version.clear();
    ip.clear();
    port = 0;
    mac.clear();
}

void tcpipInfo::parse()
{
    // Werte aufschluesseln
    int num = 0;
    unsigned char b1 = 0;
    unsigned char b2 = 0;
    unsigned char b3 = 0;
    unsigned char b4 = 0;
    char *token = std::strtok( &data[0], "\r" );
    char *pIp = nullptr;
    while ( token != nullptr )
    {
        num++;
        switch ( num )
        {
        case 1:
            version = token; 
            break;

        case 2:
            pIp = token; 
            break;

        case 3:
            try
            {
                port = boost::numeric_cast<unsigned short>( boost::lexical_cast<int>( token ) );
            }
            catch ( ... )
            {
                port = 0;
            }
            break;

        case 4:
            mac = token; 
            break;
        }

        token = std::strtok( nullptr, "\r" );
    }

    // Die IP- und Portnummern werden mit Vornullen geliefert,
    // Diese sind zu entfernen.
    num = 0;
    b1 = 0;
    b2 = 0;
    b3 = 0;
    b4 = 0;
    if ( pIp != nullptr && strlen( pIp ) )
    {
        try
        {
            token = std::strtok( pIp, "." );
            while ( token != nullptr )
            {
                num++;
                switch ( num )
                {
                case 1:
                    b1 = boost::numeric_cast<unsigned char>( boost::lexical_cast<int>( token ) );
                    break;

                case 2: 
                    b2 = boost::numeric_cast<unsigned char>( boost::lexical_cast<int>( token ) );
                    break;

                case 3: 
                    b3 = boost::numeric_cast<unsigned char>( boost::lexical_cast<int>( token ) );
                    break;

                case 4: 
                    b4 = boost::numeric_cast<unsigned char>( boost::lexical_cast<int>( token ) );
                    break;
                }

                token = std::strtok( nullptr, "." );
            }

        }
        catch ( ... )
        {
            b1 = 0;
            b2 = 0;
            b3 = 0;
            b4 = 0;
        }
    }

    // Ip ohne Vornullen
    ip = boost::str( boost::format( "%d.%d.%d.%d" ) 
        % static_cast<int>( b1 ) % static_cast<int>( b2 ) % static_cast<int>( b3 ) % static_cast<int>( b4 ) );

    struct in_addr addr;
    // IP Pruefen
#ifdef __MINGW32__
    if ( inet_addr( ip.c_str() ) == INADDR_NONE )
#else
    if ( inet_pton( AF_INET, ip.c_str(), &addr ) != 1 )
#endif
    {
        // Ungueltige Adresse oder Fehler
        ip.clear();
    }
}
